package cytoPrefuse;

import java.awt.Point;
import java.util.Arrays;

import prefuse.data.Edge;
import prefuse.data.Node;
import prefuse.visual.NodeItem;

public class SpaceGrid {
	private interface GridVisitor{
		void visitGridPoint (int x, int y, int edgePenalty, int nodeEdgePenalty, int error, Node edgeSource, Node edgeDest);
	}
	
	private class OccupyGrid implements GridVisitor{

		public void visitGridPoint(int x, int y, int edgePenalty, int nodeEdgePenalty, int error, Node edgeSource, Node edgeDest) {
			if (error == 0){
				centerGridCrossCount[x][y]++;
			}
			gridCrossCount[x][y]++;
		}	
	}
	
	private class VacateGrid implements GridVisitor{
	
		public void visitGridPoint(int x, int y,  int edgePenalty, int nodeEdgePenalty, int error, Node edgeSource, Node edgeDest) {
			if (error == 0){
				centerGridCrossCount[x][y] --;
			}
			gridCrossCount[x][y] --;
		}		
	}
	
	private class CountIntersections implements GridVisitor{
		public void visitGridPoint(int x, int y, int edgePenalty, int nodeEdgePenalty, int error, Node edgeSource, Node edgeDest) {
			int crossCount = 0;
			crossCount += gridCrossCount[x][y];
			
			if (crossCount > 0)
				crossingScore += crossCount * edgePenalty;
		 	if (error ==0 ){
				if (occupiedNodes[x][y]){
					crossingScore += nodeEdgePenalty;
				}
			}
			
		}
	
	}
	
	private boolean[][] occupiedNodes;
	private int[][] centerGridCrossCount;
	private int[][] gridCrossCount;
	
   private OccupyGrid occupyGrid = new OccupyGrid();
   private VacateGrid vacateGrid = new VacateGrid();
   private CountIntersections countIntersections = new CountIntersections();
   
   private final int FINENESS=1;
   
   private int crossingScore;
  
   private CerebralLayout layout;
   
   
   SpaceGrid (int x, int y, int maxNodes, CerebralLayout inCerebralLayout) {
		occupiedNodes = new boolean [x*FINENESS][y*FINENESS];
		centerGridCrossCount = new int [x*FINENESS][y*FINENESS];
		gridCrossCount = new int [x*FINENESS][y*FINENESS];
	
		for (int xi=0; xi <x*FINENESS; xi++){
				Arrays.fill (gridCrossCount[xi], 0);
				Arrays.fill (centerGridCrossCount[xi],0);
				Arrays.fill (occupiedNodes[xi],false);
		};
				
		layout = inCerebralLayout;
   }
   
   public void checkHealth () {
	   //For debugging
	  
	   for (int x =0; x < gridCrossCount.length; x++){
		   for (int y =0; y < gridCrossCount[x].length; y++){
			   if (gridCrossCount[x][y] < 0) {
				   System.out.println ("["+x+":"+y+"] has score " + gridCrossCount[x][y]);
			   }		
		   }
       }
	  
	  
   }
		
	
	public void addEdge (Edge e) {
			visitGridsCrossedByEdge(e, occupyGrid,0,0);	
	}
	
	public void deleteEdge (Edge e) {
			visitGridsCrossedByEdge(e, vacateGrid,0,0);
	
	}	

	private void visitGridsCrossedByEdge (Edge e, GridVisitor gv, int edgePenalty, int nodeEdgePenalty){
	    Point p0 = layout.getPositionOfNode( e.getSourceNode());
		Point p1 = layout.getPositionOfNode(e.getTargetNode());
		
		Node edgeSource = e.getSourceNode();
		Node edgeDest = e.getTargetNode();
		if (edgeSource.getRow() != edgeDest.getRow()){
			
			int x0 = p0.x * FINENESS;
			int x1 = p1.x * FINENESS;
			int y0 = p0.y *FINENESS;
			int y1 = p1.y * FINENESS;
		
			boolean steep = Math.abs(y1 - y0) > Math.abs(x1 - x0);
		    if (steep){
		    	  int temp = x0;
		    	  x0 = y0;
		    	  y0 = temp;
		    	  
		    	  temp = x1;
		    	  x1 = y1;
		    	  y1 =temp;
		    	  
		    }
		    if (x0 > x1){
		    	
		    	int temp = x0;
		    	x0 = x1;
		    	x1 = temp;
		    	
		    	temp = y0;
		    	y0 = y1;
		    	y1 = temp;
		    	
		    }
		     int deltax = x1 - x0;
		     int deltay =Math.abs(y1 - y0);
		     
		     int error = 0;
		     int y = y0;
		     int ystep = (y0 < y1)? 1 : -1;
		    
		     for (int x = x0; x <= x1; x++ ) {
		    	 if (steep)
		    		  gv.visitGridPoint(y,x, edgePenalty, nodeEdgePenalty, error, edgeSource, edgeDest);
		    	 else 
		    		  gv.visitGridPoint(x,y, edgePenalty, nodeEdgePenalty, error,edgeSource, edgeDest);
		         error = error + deltay;
		         if (2*error >= deltax) {
		             //Here's the modified version of Bresenham's.  We also visit the node we
                     //passed through as we change the y co-ordinate.
                     if (x < x1) {
                         if (error == deltax){
                             //we're exiting at an exact corner.
                             y = y + ystep;                             
                         }else if (error > deltax){
                             //we're over, check the new y immediately
                             y = y + ystep;                         
                             if (steep)
                                 gv.visitGridPoint(y,x, edgePenalty, nodeEdgePenalty, error, edgeSource, edgeDest);
                             else 
                                 gv.visitGridPoint(x,y, edgePenalty, nodeEdgePenalty, error,edgeSource, edgeDest);
                         }else{
                             //we're under, check one more x                              
                             if (steep)
                                   gv.visitGridPoint(y,x +1, edgePenalty, nodeEdgePenalty, error, edgeSource, edgeDest);
                             else 
                                   gv.visitGridPoint(x+1,y, edgePenalty, nodeEdgePenalty, error,edgeSource, edgeDest);                        
                             y = y + ystep;                         
                          }
                          error = error - deltax;
                     }
		         }
             }
		}
	}
	
	public void addNode (NodeItem n) {
		Point nPos = layout.getPositionOfNode(n);
		occupiedNodes [nPos.x][nPos.y] = true;
		
	}
	
	public void deleteNode (NodeItem n) {		
		Point nPos = layout.getPositionOfNode(n);
		occupiedNodes [nPos.x][nPos.y] = false;
	}

	
	public int countCrossings(Edge e, int edgePenalty, int nodeEdgePenalty) {
		crossingScore =0;
		visitGridsCrossedByEdge(e, countIntersections, edgePenalty, nodeEdgePenalty);
		return crossingScore;
	
	}
	
	public int countCrossings (Point p, int nodeEdgePenalty) {		
		int x = p.x;
		int y = p.y;
	
		int crossings =0;
		crossings += centerGridCrossCount[x][y] * nodeEdgePenalty;
		return crossings;
	}	
}
